/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/download.c
 * @brief code that handles the download window
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * FIXME:
 * - shutdown of gnunet-gtk does NOT terminate
 *   each of the pending downloads.  There
 *   should be a handler function that is invoked
 *   whenever gnunet-gtk shuts down and that
 *   stops all pending downloads.  This used to
 *   be implemented but got lost when Igor added
 *   the download window.
 **/

#include "gnunet_afs_esed2.h"
#include "helper.h"
#include "directorydisplay.h"
#include "download.h"
#include "main.h"

GtkWidget * dlWindow = NULL;

/* Suitable values of DownloadModel.downloadStatus */
#define DOWNLOAD_COMPLETE 0
#define DOWNLOAD_FAILED   1
#define DOWNLOAD_ABORTED  2
#define DOWNLOAD_PENDING  3

/* colors taken from x-chat source, regards */
GdkColor textColors[] = {
        {0, 0xcf3c, 0xcf3c, 0xcf3c}, /* 0  white */
        {0, 0x0000, 0x0000, 0x0000}, /* 1  black */
        {0, 0x0000, 0x0000, 0xcccc}, /* 2  blue */
        {0, 0x0000, 0xcccc, 0x0000}, /* 3  green */
        {0, 0xdddd, 0x0000, 0x0000}, /* 4  red */
        {0, 0xaaaa, 0x0000, 0x0000}, /* 5  light red */
        {0, 0xbbbb, 0x0000, 0xbbbb}, /* 6  purple */
        {0, 0xffff, 0xaaaa, 0x0000}, /* 7  orange */
        {0, 0xeeee, 0xdddd, 0x2222}, /* 8  yellow */
        {0, 0x3333, 0xdede, 0x5555}, /* 9  green */
        {0, 0x0000, 0xcccc, 0xcccc}, /* 10 aqua */
        {0, 0x3333, 0xeeee, 0xffff}, /* 11 light aqua */
        {0, 0x0000, 0x0000, 0xffff}, /* 12 blue */
        {0, 0xeeee, 0x2222, 0xeeee}, /* 13 light purple */
        {0, 0x7777, 0x7777, 0x7777}, /* 14 grey */
        {0, 0x9999, 0x9999, 0x9999}, /* 15 light grey */
        {0, 0xa4a4, 0xdfdf, 0xffff}, /* 16 marktext Back (blue) */
        {0, 0x0000, 0x0000, 0x0000}, /* 17 marktext Fore (black) */
        {0, 0xdf3c, 0xdf3c, 0xdf3c}, /* 18 foreground (white) */
        {0, 0x0000, 0x0000, 0x0000}, /* 19 background (black) */
        {0, 0x8c8c, 0x1010, 0x1010}, /* 20 tab New Data (dark red) */
        {0, 0x0000, 0x0000, 0xffff}, /* 21 tab Nick Mentioned (blue) */
        {0, 0xf5f5, 0x0000, 0x0000}, /* 22 tab New Message (red) */
};

/**
 * @brief state associated with a GTK download window
 **/
typedef struct {
  RootNode root;
  char * fileName;
  pthread_t downloadThread;
  RequestManager * rm;
  Semaphore * doneSem;
  cron_t downloadStartTime;
  int successfulStart;
  int downloadStatus;
  TIME_T lastDisplayTime;
  GtkWidget * dlList;
} DownloadModel;

/**
 * "delete_event" handler for a download. Sets the 
 * download state as aborted, and releases the wait semaphore for 
 * the thread awaiting download completion. Returns TRUE so that 
 * gtk doesn't destroy the widget (its never destroyed).
 *
 * @param widget not used
 * @param clist the list of downloads
 **/
static gint abortSelectedDownloads(GtkWidget * widget,
			           GtkWidget * clist) {
  DownloadModel * dlm;
  int row;
  GList * tmp;

  LOG(LOG_DEBUG,
      "DEBUG: abortSelectedDownloads %x (%x)\n",
      clist, widget);
  if(!GTK_CLIST(clist))
    return TRUE;
  
  gtk_clist_freeze(GTK_CLIST(clist));

  tmp = GTK_CLIST(clist)->selection;
  while (tmp) {
    row = (int)tmp->data;
    tmp = tmp->next;

    /* if dlm is not NULL, abort the download */
    dlm = gtk_clist_get_row_data(GTK_CLIST(clist),
	   	  	         row);
    if(dlm) {
      dlm->downloadStatus = DOWNLOAD_ABORTED;
      SEMAPHORE_UP(dlm->doneSem);
    } 
    
    /* remove list entry */
    gtk_clist_remove(GTK_CLIST(clist),
		     row);
  }

  gtk_clist_thaw(GTK_CLIST(clist));
  
  return TRUE;
}


/**
 * This method is called by the download code to notify the user
 * interface of the download progress.
 *
 * @param stats the new statistical values
 * @param dlm the accessor to the GTK window
 **/
static void modelCallback(ProgressStats * stats,
			  DownloadModel * dlm) {
  double currentRetryAvg, averageBps, percentage;
  char pos[16], kbs[14], perc[14], areq[14], cra[14], tr[14];
  gint row;
  TIME_T now;

  if (dlm->downloadStatus != DOWNLOAD_PENDING) 
    return;

  /* don't display more often than once/sec */
  TIME(&now);
  if ((now-(dlm->lastDisplayTime)) < 1 &&
      (stats->filesize != stats->progress))
    return;
  else
    dlm->lastDisplayTime = now;
 
  if(stats->requestsSent > 0)
    currentRetryAvg = 
      (double)stats->currentRetries / 
      (double)stats->requestsSent;
  else
    currentRetryAvg = 0;
  
  if (cronTime(NULL) - dlm->downloadStartTime > 0)
    averageBps =
      (double)(stats->progress) / 
      (((double)(cronTime(NULL) - dlm->downloadStartTime))/(double)cronSECONDS);
  else
    averageBps = 0;

  if(stats->filesize>0)
    percentage = 100.0*((double)stats->progress/(double)stats->filesize);
  else
    percentage = 0;

  snprintf(pos, sizeof(pos), "%d", stats->progress);
  snprintf(kbs, sizeof(kbs), "%.1f", averageBps);
  snprintf(perc, sizeof(perc), "%3.1f%%", percentage);
  snprintf(areq, sizeof(areq), "%d", stats->requestsSent);
  snprintf(cra, sizeof(cra), "%3.1f", currentRetryAvg);
  snprintf(tr, sizeof(tr), "%d", stats->totalRetries);
 
  gdk_threads_enter();
  gtk_clist_freeze(GTK_CLIST(dlm->dlList));
  row = gtk_clist_find_row_from_data(GTK_CLIST(dlm->dlList),
				     dlm);
  gtk_clist_set_text(GTK_CLIST(dlm->dlList),
		     row,
		     1,
		     perc);
  gtk_clist_set_text(GTK_CLIST(dlm->dlList),
		     row,
		     2,
		     pos);
  gtk_clist_set_text(GTK_CLIST(dlm->dlList),
		     row,
		     4,
		     areq);
  gtk_clist_set_text(GTK_CLIST(dlm->dlList),
		     row,
		     5,
		     cra);
  gtk_clist_set_text(GTK_CLIST(dlm->dlList),
		     row,
		     6,
		     tr);
  gtk_clist_set_text(GTK_CLIST(dlm->dlList),
		     row,
		     7,
		     kbs);
  if(dlm->successfulStart == NO && stats->progress > 0) {
      gtk_clist_set_foreground(GTK_CLIST(dlm->dlList), 
                               row,
                               &textColors[1]);
      dlm->successfulStart = YES;
  } 
  gtk_clist_thaw(GTK_CLIST(dlm->dlList));
  gdk_threads_leave();

  if (stats->filesize == stats->progress) {
    /* reset the request counters (just cosmetic) */
    snprintf(areq, sizeof(areq), "0");
    snprintf(cra, sizeof(cra), "0.0");
    
    gdk_threads_enter();
    gtk_clist_freeze(GTK_CLIST(dlm->dlList));
    gtk_clist_set_text(GTK_CLIST(dlm->dlList),
 	  	       row,
		       4,
		       areq);
    gtk_clist_set_text(GTK_CLIST(dlm->dlList),
  	  	       row,
		       5,
		       cra);
    gtk_clist_thaw(GTK_CLIST(dlm->dlList));
    refreshMenuSensitivity();
    gdk_threads_leave();

    if (stats->filesize == 0)       
      dlm->downloadStatus = DOWNLOAD_FAILED;
    else
      dlm->downloadStatus = DOWNLOAD_COMPLETE;
    SEMAPHORE_UP(dlm->doneSem); /* signal: we're done with download */
    return;
  }
}

/**
 * Main function of the download thread. This function terminates
 * automagically once the download is complete or if the download
 * entry is removed while dl.  It is responsible for stopping the
 * requestmanager.
 * 
 * @param dlm the download model
 **/
static void downloadFile_(DownloadModel * dlm) {
  char * mime;
  gint row;

  LOG(LOG_DEBUG,
      "DEBUG: Entering downloadFile_ for %s (%x)\n", 
      dlm->fileName,
      dlm);

  /* initiate download */
  cronTime(&dlm->downloadStartTime);
  
  /* this starts the "real" download thread in the background,
     with "modelCallback" called back to tell us about the progress */
  dlm->rm = downloadFile(&dlm->root.header.fileIdentifier,
			 dlm->fileName,
			 (ProgressModel)&modelCallback,
			 dlm);

  /* Wait here until download is complete or
     the window is closed or gnunet-gtk is terminated */
  LOG(LOG_DEBUG,
      "DEBUG: waiting for DL completion (%x)\n",
      dlm);
  SEMAPHORE_DOWN(dlm->doneSem);
  LOG(LOG_DEBUG,
      "DEBUG: download complete (%d) enter destroyRequestManager (%x)\n",
      dlm->downloadStatus,
      dlm);

  /* stop the RequestManager */
  destroyRequestManager(dlm->rm); 

  /*
    ok, now why are we here? 4 possibilities:
     a) download aborted (user closed window)
     b) gnunet-gtk terminated (same as download aborted)
     c) download failed (gnunetd exit, out-of-space)
     d) download completed 

     In case "d" we show the "YAY" window
     and wait for another signal.
     In case "c" we show the "BAH" window
     and wait for another signal.
  */
  switch (dlm->downloadStatus) {
  case DOWNLOAD_COMPLETE:
    gdk_threads_enter();
    /* color successful dl green */ 
    gtk_clist_freeze(GTK_CLIST(dlm->dlList));
    row = gtk_clist_find_row_from_data(GTK_CLIST(dlm->dlList),
				       dlm);
    gtk_clist_set_foreground(GTK_CLIST(dlm->dlList), 
                             row,
                             &textColors[3]);
    gtk_clist_set_text(GTK_CLIST(dlm->dlList),
  	  	       row,
		       1,
		       "DONE");
    gtk_clist_thaw(GTK_CLIST(dlm->dlList));
    
    switch (ntohs(dlm->root.header.major_formatVersion)) {
    case ROOT_MAJOR_VERSION:
      mime = dlm->root.header.mimetype;
      break;
    case SBLOCK_MAJOR_VERSION:
      mime = ((SBlock*)&dlm->root)->mimetype;
      break;
    default:
      mime = "unknown";
      break;
    }
    if (0 == strcmp(mime,
		    GNUNET_DIRECTORY_MIME)) {
      displayDirectory(dlm->fileName,
		       &dlm->root);
    }
    gdk_threads_leave();
    SEMAPHORE_DOWN(dlm->doneSem); /* wait for window closing */
    break;
  case DOWNLOAD_FAILED:
    gdk_threads_enter(); 
    /* color failed dl red */ 
    gtk_clist_freeze(GTK_CLIST(dlm->dlList));
    row = gtk_clist_find_row_from_data(GTK_CLIST(dlm->dlList),
				       dlm);
    gtk_clist_set_foreground(GTK_CLIST(dlm->dlList), 
                             row,
                             &textColors[4]);
    gtk_clist_set_text(GTK_CLIST(dlm->dlList),
  	  	       row,
		       1,
		       "FAIL");
    gtk_clist_thaw(GTK_CLIST(dlm->dlList));
    gdk_threads_leave();
    SEMAPHORE_DOWN(dlm->doneSem); /* wait for window closing */    
    break;
  default:
    /* do nothing */
    break;
  }

  /* finally, disentangle from the clist and free dlm resources */
  gdk_threads_enter();
  gtk_clist_freeze(GTK_CLIST(dlm->dlList));
  row = gtk_clist_find_row_from_data(GTK_CLIST(dlm->dlList),
				     dlm);
  gtk_clist_set_row_data(GTK_CLIST(dlm->dlList),
                         row,
			 NULL);
  gtk_clist_thaw(GTK_CLIST(dlm->dlList));
  gdk_threads_leave();
  SEMAPHORE_FREE(dlm->doneSem);
  FREE(dlm->fileName);
  FREE(dlm);
}

/**
 * Open the download window and start the download of a file in the
 * background. The method executes during a signal handler, so a GTK
 * lock is not required to to GUI operations.
 *
 * @param filename the name of the file to download
 *        (must copy, will be freed by caller)
 * @param root information about what to download 
 *        (will be freed by caller, must copy!)
 **/
void startDownload(gchar * filename,
		   RootNode * root) {
  GtkWidget * clist;
  char * fileNameRoot;
  int i;
  gint row;
  DownloadModel * dlm;
  gchar * fileInfo[8];

  dlm = MALLOC(sizeof(DownloadModel));
  memset(dlm, 
	 0, 
	 sizeof(DownloadModel));
  dlm->fileName = STRDUP(filename);
  memcpy(&dlm->root,
	 root,
	 sizeof(RootNode));

  fileNameRoot = dlm->fileName;
  for (i=strlen(dlm->fileName)-1;i>=0;i--) {
    if (dlm->fileName[i] == '/') {
      fileNameRoot = &dlm->fileName[i+1];
      break;
    }
  }
  
  /* create new download window? */
  if (!dlWindow) {
    GtkWidget * scrolled_window;
    GtkWidget * button;
    GtkWidget * box;
    GtkWidget * entry;
    static gchar * descriptions[] = {
      "Filename",
      "%",              /* Completion percentage */
      "Position",	/* Bytes dl'ed so far */
      "Size",
      "AReq",		/* Active block requests */
      "CR/A",           /* Current Retries per Active requests */
      "totR",           /* Total Retries */
      "BPS",            /* Current bytes per second -estimate */
    };
    static int widths[] = {
      290, 50, 70, 70, 40,30,50,50
    };

    dlWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(dlWindow), 
  	  	         "gnunet-gtk: Downloads");
    gtk_widget_set_usize(GTK_WIDGET(dlWindow),
                         780, /* x-size */
                         300); /* y-size */
    gtk_signal_connect_object(GTK_OBJECT(dlWindow), "delete_event",
                              GTK_SIGNAL_FUNC(hideWindow),
                              GTK_OBJECT(dlWindow));
    gtk_signal_connect_object(GTK_OBJECT(dlWindow), "destroy",
                              GTK_SIGNAL_FUNC(hideWindow),
                              GTK_OBJECT(dlWindow));


    box = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(dlWindow), 
  	  	      box);
    gtk_container_set_border_width(GTK_CONTAINER(dlWindow), 
 				   8);
   
    /* scrolled window for the dl list */
    scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
                                   GTK_POLICY_AUTOMATIC,
                                   GTK_POLICY_ALWAYS);
    gtk_box_pack_start(GTK_BOX(box),
                       scrolled_window,
                       TRUE,
                       TRUE,
                       0);
    gtk_widget_show(scrolled_window);
    /* create a list to hold the downloads in */
    clist = gtk_clist_new_with_titles(8, descriptions);
    gtk_clist_set_selection_mode
              (GTK_CLIST(clist),
	       GTK_SELECTION_EXTENDED);
    for (i=0;i<8;i++)
      gtk_clist_set_column_width(GTK_CLIST(clist),
                                 i,
                                 widths[i]);
    gtk_container_add(GTK_CONTAINER(scrolled_window),
    		      clist);
    gtk_object_set_data(GTK_OBJECT(dlWindow),
    			"LIST",
			clist);
    gtk_widget_show(box);
    gtk_widget_show(scrolled_window);
    /* cancel/remove button */
    button = gtk_button_new_with_label("Remove selected entries");
    gtk_signal_connect (GTK_OBJECT(button),
                      "clicked",
                      GTK_SIGNAL_FUNC(abortSelectedDownloads),
                      clist);
    gtk_box_pack_start(GTK_BOX(box),
                       button,
                       FALSE,
                       FALSE,
                       0);
    gtk_widget_show(button);

    /* activate the pulldown menu option now */
    entry = gtk_item_factory_get_widget(itemFactory,
                                        "/File/Show downloads");
    gtk_widget_set_sensitive(entry, TRUE);
  } else {
    /* use existing dlWindow and list */
    clist = gtk_object_get_data(GTK_OBJECT(dlWindow),
    				"LIST");
  }

  /* set default disp. values for new download */
  fileInfo[0]=STRDUP(fileNameRoot);
  fileInfo[1]="0%";
  fileInfo[2]="-";
  fileInfo[3]=MALLOC(32);
  fileInfo[4]="-";
  fileInfo[5]="-";
  fileInfo[6]="-";
  fileInfo[7]="-";
  snprintf(fileInfo[3], 
	   32, 
	   "%u", 
	   (unsigned int) ntohl(dlm->root.header.fileIdentifier.file_length));

  TIME(&dlm->lastDisplayTime);

  gtk_clist_freeze(GTK_CLIST(clist));  
  row = gtk_clist_append(GTK_CLIST(clist),
			 fileInfo);
  FREE(fileInfo[0]);
  FREE(fileInfo[3]);
  gtk_clist_set_foreground(GTK_CLIST(clist), 
			   row,
                          &textColors[11]);
  gtk_clist_set_row_data(GTK_CLIST(clist),
  			 row,
			 dlm);
  gtk_clist_thaw(GTK_CLIST(clist));
			 
  /* remember clist for updates in the DLM struct */
  dlm->dlList = clist;
  dlm->downloadStatus = DOWNLOAD_PENDING;
  dlm->doneSem = SEMAPHORE_NEW(0);
  
  /* display download window */
  gtk_widget_show(clist);
  gtk_widget_show(dlWindow);

  /* create thread that runs the download */
  if (0 != PTHREAD_CREATE(&dlm->downloadThread,
			  (PThreadMain) &downloadFile_,
			  dlm,
			  64 * 1024))
    errexit("FATAL: could not create download thread (%s)!\n",
	    strerror(errno));
  pthread_detach(dlm->downloadThread);  
}

/**
 * Starts a file download when user has filled in the fields
 **/ 
void fetchURICallback(GtkWidget * widget,
			 gpointer data) {
  gchar * uri;
  FileIdentifier * fid;
  GtkWidget * entry;
  RootNode root;

  entry = gtk_object_get_data(GTK_OBJECT(data),
			      "entry");

  uri = gtk_entry_get_text(GTK_ENTRY(entry));
  if(!uri)
    return;

  fid = stringToFileIdentifier(uri);
  if(!fid) {
    guiMessage("Invalid gnunet AFS URI");
    gtk_widget_destroy(data);
    return;
  }

  root.header.major_formatVersion = ROOT_MAJOR_VERSION;
  root.header.minor_formatVersion = ROOT_MINOR_VERSION;
  memcpy(&root.header.fileIdentifier,
	 fid,
         sizeof(FileIdentifier));
  strcpy(root.header.mimetype,
	 "unknown");
  
  /* FIXME: Future URLs should contain suggested filename 
   * (or the code should ask one from user) */
  sprintf(root.header.filename, "unknown.%ld", (unsigned long)time(NULL));

  startDownload(root.header.filename,
                &root);

  gtk_widget_destroy(data);
}

void fetchURI(GtkWidget * widget,
	      gpointer data) {
  GtkWidget * window;
  GtkWidget * vbox;
  GtkWidget * label;
  GtkWidget * entry;
  GtkWidget * hbox;
  GtkWidget * button_ok;
  GtkWidget * button_cancel;

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_widget_set_usize(GTK_WIDGET(window),
  		       780,
		       100);
  gtk_window_set_title(GTK_WINDOW(window),
  		       "Download URI");
  /* add container for window elements */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window),
                    vbox);
  gtk_widget_show(vbox);

  /* when user clicks on close box, always "destroy" */
  gtk_signal_connect(GTK_OBJECT(window),
                     "destroy",
                     GTK_SIGNAL_FUNC(destroyWidget),
                     window);

  gtk_container_set_border_width(GTK_CONTAINER(window),
                                 10);
  
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
		     hbox,
		     FALSE,
		     TRUE,
		     0);
  gtk_widget_show(hbox);
 
  label = gtk_label_new("GNUnet AFS URI: ");
  gtk_box_pack_start(GTK_BOX(hbox),
		     label,
		     FALSE,
		     FALSE,
		     0);
  gtk_widget_show(label);
  entry = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(hbox),
		     entry,
                     TRUE,
		     TRUE,	
		     0);
  gtk_signal_connect(GTK_OBJECT(entry),
                     "activate",
                     GTK_SIGNAL_FUNC(fetchURICallback),
                     window);
  gtk_object_set_data(GTK_OBJECT(window),
		      "entry",
		      entry);
  gtk_widget_show(entry);

  button_ok = gtk_button_new_with_label("Ok");
  button_cancel = gtk_button_new_with_label("Cancel");

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
                     hbox,
                     FALSE,
                     TRUE,
                     0);
  gtk_box_pack_start(GTK_BOX(hbox),
		     button_ok,
		     TRUE,
		     TRUE,
		     0);
  gtk_box_pack_start(GTK_BOX(hbox),
		     button_cancel,
		     TRUE,
		     TRUE,
		     0);
  gtk_signal_connect(GTK_OBJECT(button_cancel),
		     "clicked",
		     GTK_SIGNAL_FUNC(destroyWidget),
		     window);
  gtk_signal_connect(GTK_OBJECT(button_ok),
		     "clicked",
		     GTK_SIGNAL_FUNC(fetchURICallback),
		     window);
  gtk_widget_show(hbox);
  gtk_widget_show(button_ok);
  gtk_widget_show(button_cancel);
  gtk_widget_show(window);
 


}
		 

/* end of download.c */
