/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file src/applications/afs/gtkui/helper.c
 * @brief This file contains some GUI helper functions
 * @author Igor Wronsky
 **/

#include "gnunet_afs_esed2.h"
#include "helper.h"
#include <stdlib.h>
#include "main.h"

#define HELPER_DEBUG NO

GtkWidget * infoWindow = NULL;
static GtkWidget * infoText = NULL;

/**
 * Callback for handling "delete_event": close the window 
 **/
gint deleteEvent(GtkWidget * widget,
		 GdkEvent * event,
		 gpointer data) {
#if DEBUG_HELPER
  LOG(LOG_DEBUG, 
      "DEBUG: deleteEvent\n");
#endif
  return FALSE;
}

/**
 * A callback to destroy any widget given as second argument
 **/
void destroyWidget(GtkWidget * dummy, 
		   GtkWidget * widget) {
#if DEBUG_HELPER
  LOG(LOG_DEBUG, 
      "DEBUG: destroyWidget %d\n", 
      widget);
#endif
  gtk_widget_destroy(widget);
}

/** 
 * Displays an informative message to the user in a fresh window 
 **/
void guiMessage(const char * format, ...) {
  va_list args;
  gchar *note;
  GtkWidget * window;
  GtkWidget * label;
  GtkWidget * box;
  GtkWidget * button;

  va_start(args, format);
  note = g_strdup_vprintf(format, args);
  va_end(args);

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_container_set_border_width(GTK_CONTAINER(window), 10);
  gtk_window_set_policy(GTK_WINDOW(window),
			FALSE,
			FALSE,
			TRUE);
  gtk_window_set_position(GTK_WINDOW(window),
			  GTK_WIN_POS_MOUSE);
  gtk_window_set_title(GTK_WINDOW(window), 
		       "Note");
  gtk_signal_connect(GTK_OBJECT(window), 
		     "delete_event",
		     GTK_SIGNAL_FUNC(deleteEvent), 
		     NULL);

  box = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), 
		    box);
  gtk_widget_show(box);

  label = gtk_label_new(note);
  free(note); /* allocated in g_strdup_vprintf */
  gtk_box_pack_start(GTK_BOX(box),
		     label,
		     FALSE,
		     FALSE,
		     0);
  gtk_widget_show(label);
  
  button = gtk_button_new_with_label("Ok");
  gtk_signal_connect(GTK_OBJECT (button),
		      "clicked",
		      GTK_SIGNAL_FUNC(destroyWidget),
		     window);
  gtk_box_pack_start(GTK_BOX(box),button,FALSE,FALSE,0);
  gtk_widget_show(button);
  
  gtk_widget_show(window);
}

/** 
 * Appends a message to the info window
 **/
void infoMessage(const char * format, ...) {
  va_list args;
  gchar *note;

  va_start(args, format);
  note = g_strdup_vprintf(format, args);
  va_end(args);

  if(!infoWindow) {
    GtkWidget * box1;
    GtkWidget * button;
    GtkWidget * scrolled_window;

    infoWindow = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect(GTK_OBJECT(infoWindow),
                       "delete_event",
                       GTK_SIGNAL_FUNC(deleteEvent),
                       NULL);

    gtk_window_set_title(GTK_WINDOW(infoWindow),
                         "Information");
    gtk_widget_set_usize(GTK_WIDGET(infoWindow),
                         780,
                         300);

    box1 = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER (infoWindow),
                      box1);
    gtk_widget_show(box1);
    
    /* create a scrollable window */
    scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
    				   GTK_POLICY_AUTOMATIC,
				   GTK_POLICY_ALWAYS);
    gtk_box_pack_start(GTK_BOX(box1),
    		       scrolled_window,
		       TRUE,
		       TRUE,
		       0);
    gtk_widget_show(scrolled_window);

    /* create a text widget */
    infoText = gtk_text_new(NULL, NULL);
    gtk_text_set_editable(GTK_TEXT (infoText),
                          FALSE);
    gtk_container_add(GTK_CONTAINER(scrolled_window),
    		      infoText);
    gtk_widget_show(infoText);
    gtk_widget_realize(infoText);
  
    /* finish with a close button */
    button = gtk_button_new_with_label("Close");
    gtk_box_pack_start(GTK_BOX (box1),
                       button,
                       FALSE,
                       FALSE,
                       0);
    gtk_signal_connect_object(GTK_OBJECT(button),
                              "clicked",
                              GTK_SIGNAL_FUNC(hideWindow),
                              GTK_OBJECT(infoWindow));
    gtk_signal_connect_object(GTK_OBJECT(infoWindow), "delete_event",
                              GTK_SIGNAL_FUNC(hideWindow),
                              GTK_OBJECT(infoWindow));
    gtk_signal_connect_object(GTK_OBJECT(infoWindow), "destroy",
                              GTK_SIGNAL_FUNC(hideWindow),
                              GTK_OBJECT(infoWindow));
    gtk_widget_show(button);
  }
  gtk_widget_show(infoWindow);

  /* append the text */
  gtk_text_freeze(GTK_TEXT(infoText));
  gtk_text_insert(GTK_TEXT(infoText),
                  NULL,
                  &infoText->style->black,
                  NULL,
                  note, -1);
  gtk_text_thaw(GTK_TEXT(infoText));


}


GtkNotebook * notebook = NULL;

void addToNotebook(char * labelName,
		   GtkWidget * frame) {
  GtkWidget * entry;

  /* add a new notebook for the search results */
  GtkWidget * label = gtk_label_new(labelName);
  gtk_notebook_append_page(notebook, frame, label);
  gtk_widget_show(frame);

  /* activate menus */
  entry = gtk_item_factory_get_widget(itemFactory,
                                      "/File/Select all");
  gtk_widget_set_sensitive(entry, TRUE);
  entry = gtk_item_factory_get_widget(itemFactory,
                                      "/File/Select none");
  gtk_widget_set_sensitive(entry, TRUE);
  entry = gtk_item_factory_get_widget(itemFactory,
                                      "/File/Abort search");
  gtk_widget_set_sensitive(entry, TRUE);
}

void hideWindow(GtkWidget * widget,
		gpointer data) {
  if(widget)
    gtk_widget_hide(widget);
}

/**
 * A dirty way to pump some gnunetd stats to the info window
 **/
void showStats(GtkWidget * widget,
	       gpointer data) {
  FILE *fp;
  char buffer[512];
  char fn[512];
 
  infoMessage("### [ gnunetd stats ] ################################################\n");

  sprintf(fn, "gnunet-gtk_stats.XXXXXX");
  mkstemp(fn);

  sprintf(buffer, "gnunet-stats >%s", fn);
  system(buffer);
  fp=fopen(fn, "r");
  if(!fp) {
    LOG(LOG_ERROR, "Unable to open %s for reading\n", 
    	fn);
    return;
  }
  while(!feof(fp)) {
    fgets(buffer, 512, fp);
    infoMessage(buffer);
  }
  fclose(fp);
  unlink(fn);
  
  infoMessage("#################################################################\n");
}

/** 
 * Checks if gnunetd is running
 * 
 * NOTE: Uses a potentially nasty/lazy way to check if gnunetd
 * is running (sends a stat req message but doesn't read a reply).
 **/
static int checkDaemonRunning(void) {
  GNUNET_TCP_SOCKET * sock;
  CS_HEADER csHdr;

  sock = getClientSocket();
  if(sock == NULL) {
    LOG(LOG_WARNING, "WARNING: socket create failed, shouldn't happen\n");
    return SYSERR;  
  }    

  csHdr.size
    = htons(sizeof(CS_HEADER));
  csHdr.tcpType
    = htons(STATS_CS_PROTO_GET_STATISTICS);
  if (SYSERR == writeToSocket(sock,
                              &csHdr)) {
    LOG(LOG_DEBUG, "DEBUG: gnunetd is NOT running\n");
    releaseClientSocket(sock);
    return SYSERR;
  } else {
    LOG(LOG_DEBUG, "DEBUG: gnunetd is running\n");
    releaseClientSocket(sock);
    return OK;
  }
}

/** 
 * Launch gnunetd, don't check if its running
 **/
static void launchDaemonNoCheck(GtkWidget * widget,
 	          gpointer data) {
  system("nice gnunetd");
  guiMessage("Launched gnunetd...");
  gtk_widget_destroy(GTK_WIDGET(data));
}

/** 
 * Launch gnunetd w/ checks
 **/
void launchDaemon(GtkWidget * widget,
 	          gpointer data) {
  if(OK == checkDaemonRunning() ) {
    guiMessage("gnunetd is already running");
    return;
  } else {
    system("nice gnunetd");
    guiMessage("Launched gnunetd...");
  }
}

/** 
 * Kill gnunetd
 **/
void killDaemon(GtkWidget * widget,
 	        gpointer data) {
  if(OK == checkDaemonRunning() ) {
    system("killall gnunetd");
    guiMessage("Executed 'killall gnunetd'...");
  } else {
    guiMessage("gnunetd is not running...");
  }
}

/**
 * Ask if the user wishes to start gnunetd
 **/
static void initDaemonStartDialog(void) {
   GtkWidget *dialog;
   GtkWidget *label;
   GtkWidget *okay_button;
   GtkWidget *no_button;

   dialog = gtk_dialog_new();
   label = gtk_label_new("gnunetd (daemon) doesn't seem to be running.\nWould you like to start it?\n");
   gtk_container_add (GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
                      label);

   okay_button = gtk_button_new_with_label("Yes!");
   no_button = gtk_button_new_with_label("Naah");
   
   gtk_signal_connect(GTK_OBJECT(okay_button), 
                      "clicked",
                      GTK_SIGNAL_FUNC(launchDaemonNoCheck), 
		      dialog);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
                      okay_button);
   gtk_signal_connect(GTK_OBJECT(no_button), 
		      "clicked",
                      GTK_SIGNAL_FUNC(destroyWidget), 
                      dialog);
   gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->action_area),
                      no_button);

   gtk_widget_show_all(dialog);
}


/**
 * Checks if gnunetd is running and if not, prompts user 
 * to run gnunetd. Always returns OK.
 *
 **/
int checkForDaemon(void) {
  if(SYSERR == checkDaemonRunning()) {
    initDaemonStartDialog();
  }

  return OK;
}

/* end of helper.c */

