/*
     This file is part of GNUnet.
     (C) 2002, 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/tools/gnunet-directory-print.c
 * @brief tool to print the contents of a GNUnet directory file
 * @author Christian Grothoff
 **/

#include "gnunet_afs_esed2.h"
 
static char ** filenames;
static int filenamescnt;

static void printNode(RootNode * root,
		      void * unused) {
  char * string;

  makeRootNodeAvailable(root,
			DIR_CONTEXT_DIRECTORY);
  string = rootNodeToString(root);
  printf("%s\n", string);
  FREE(string);
}

static void printDirectory(char * filename) {
  GNUnetDirectory * dir;
  unsigned int i;

  filename = expandFileName(filename);
  printf("==> Directory '%s':\n",
	 filename);
  dir = readGNUnetDirectory(filename);
  if (dir == NULL) { 
    printf("ERROR: no such file or invalid format for GNUnet directory.\n");
    return;
  }
  dir->description[MAX_DESC_LEN-1] = '\0';
  printf("%s\n",
	 dir->description);
  for (i=0;i<ntohl(dir->number_of_files);i++) {
    printf("%u: ", i);
    printNode(&dir->contents[i], NULL);
  }     
  printf("\n");
  FREE(dir);
  FREE(filename);
}

/**
 * Print a list of the options we offer.
 **/
static void printhelp() {
  printf("USAGE: gnunet-directory-print FILENAME(s)\n");
  printHelpDefault();
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseCommandLine(int argc, 
			    char * argv[]) {
  int shouldexit=0;
  int c;

  FREENONNULL(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL));
  while (1) {
    int option_index = 0;
    static struct GNoption long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { 0,0,0,0 }
    };
    
    c = GNgetopt_long(argc,
		      argv, 
		      "vhdc:L:", 
		      long_options, 
		      &option_index);    
    if (c == -1) 
      break;  /* No more flags to process */
    if (YES == parseDefaultOptions(c, GNoptarg))
      continue;    
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-directory-print v%s\n",
	     VERSION, 
	     AFS_VERSION);
      shouldexit=1;
      break;
    case 'h': 
      printhelp(); 
      shouldexit=1;
      break;
    default:
      printf("Unknown option %c. Aborting.\n"
	     "Use --help to get a list of options.\n",
	     c);
      shouldexit=1;    
    } /* end of parsing commandline */
  }
  filenames = &argv[GNoptind];
  filenamescnt = argc - GNoptind;
  if (filenamescnt == 0) {
    printf("ERROR: you must specify a filename.\n");
    return 1;
  }
  return shouldexit;
}

int main(int argc, 
	 char * argv[]) {
  int i;
  if (SYSERR == initUtil(argc, argv, &parseCommandLine))
    return 1;

  for (i=0;i<filenamescnt;i++)
    printDirectory(filenames[i]);
  doneUtil();
  return 0;
}


/* end of gnunet-directory-listdb.c */
