/*
     This file is part of GNUnet.
     (C) 2003 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/io.c
 * @brief (network) input/output operations
 * @author Christian Grothoff
 **/

#include "gnunet_util.h"

/* some systems send us signals, so we'd better
   catch them (& ignore) */
#ifndef LINUX
static void catcher(int sig) {
  LOG(LOG_INFO,
      "INFO: signal %d caught\n", 
      sig);
  /* re-install signal handler! */
  signal(sig, catcher);
}


#endif

void gnunet_util_initIO() {
#ifndef LINUX
  if ( SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
    if ( SIG_ERR == signal(SIGPIPE, catcher))
      LOG(LOG_WARNING,
	  "WARNING: could not install handler for SIGPIPE!\n"\
	  "Attempting to continue anyway.");
  }
#endif
}

void gnunet_util_doneIO() {
}


/* recv wrappers */

/**
 * Do a NONBLOCKING read on the given socket.  Note that in order to
 * avoid blocking, the caller MUST have done a select call before
 * calling this function.
 *
 * Reads at most max bytes to buf.  On error, return SYSERR (errors
 * are blocking or invalid socket but NOT a partial read).  Interrupts
 * are IGNORED.
 *
 * @return the number of bytes read or SYSERR. 
 *         0 is returned if no more bytes can be read
 */ 
int RECV_NONBLOCKING(int s,
		     void * buf,
		     size_t max) {  
  int ret;
  
  do {
#ifdef CYGWIN
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = recv(s,
	       buf,
	       max,
	       MSG_NOSIGNAL);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#elif OSX
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = recv(s,
	       buf,
	       max,
	       0);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#elif SOMEBSD
    ret = recv(s,
	       buf,
	       max,
	       MSG_DONTWAIT);
#elif SOLARIS
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = recv(s,
	       buf,
	       max,
	       MSG_DONTWAIT);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#elif LINUX
   ret = recv(s,
	       buf,
	       max,
	       MSG_DONTWAIT|MSG_NOSIGNAL);
#else
    /* good luck */
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = recv(s,
	       buf,
	       max,
	       0);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#endif
  } while ( ( ret == -1) && ( errno == EINTR) );
  if ( (ret < 0) || (ret > max) )
    return SYSERR;
  return ret;
}

/**
 * Do a BLOCKING read on the given socket.  Read len bytes (if needed
 * try multiple reads).  Interrupts are ignored.
 *
 * @return SYSERR if len bytes could not be read,
 *   otherwise the number of bytes read (must be len)
 */
int RECV_BLOCKING_ALL(int s,
		      void * buf,
		      size_t len) {
  size_t pos;
  int i;

  pos = 0;
  while (pos < len) {
#ifdef LINUX
    i = recv(s,
	     &((char*)buf)[pos],
	     len - pos,
	     MSG_NOSIGNAL);
#elif CYGWIN
    i = recv(s,
	     &((char*)buf)[pos],
	     len - pos,
	     MSG_NOSIGNAL);
#else
    i = recv(s,
	     &((char*)buf)[pos],
	     len - pos,
	     0);
#endif
    if ( (i == -1) && (errno == EINTR) )
      continue;
    if (i <= 0)
      return SYSERR;
    pos += i;
  }
  if (pos != len)
    errexit("ASSERTION failed: %u != %u\n",
	    len, pos);
  return pos;
}

/**
 * Do a NONBLOCKING write on the given socket.
 * Write at most max bytes from buf.  On error,
 * return SYSERR (errors are blocking or invalid
 * socket but NOT an interrupt or partial write).
 * Interrupts are ignored (cause a re-try).
 *
 * @return the number of bytes written or SYSERR. 
 */ 
int SEND_NONBLOCKING(int s,
		     void * buf,
		     size_t max) {  
  int ret;

  do {
#ifdef SOMEBSD
    ret = send(s,
	       buf,
	       max,
	       MSG_DONTWAIT);
#elif SOLARIS
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = send(s,
	       buf,
	       max,
	       MSG_DONTWAIT);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#elif OSX
    /* As braindead as Win32? */
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = send(s,
	       buf,
	       max,
	       0);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#elif CYGWIN
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = send(s,
	       buf,
	       max,
	       MSG_NOSIGNAL);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);
#elif LINUX
    ret = send(s,
	       buf,
	       max,
	       MSG_DONTWAIT|MSG_NOSIGNAL);
#elif
    /* pray */
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) | O_NONBLOCK);
    ret = send(s,
	       buf,
	       max,
	       0);
    fcntl(s, 
	  F_SETFL, 
	  fcntl(s, F_GETFL) & ~O_NONBLOCK);   
#endif
  } while ( (ret == -1) &&
	    (errno == EINTR) );
  if ( (ret < 0) || (ret > max) )
    return SYSERR;
  return ret;
}

/**
 * Do a BLOCKING write on the given socket.  Write len bytes (if
 * needed do multiple write).  Interrupts are ignored (cause a
 * re-try).
 *
 * @return SYSERR if len bytes could not be send,
 *   otherwise the number of bytes transmitted (must be len)
 */
int SEND_BLOCKING_ALL(int s,
		      void * buf,
		      size_t len) {
  size_t pos;
  int i;

  pos = 0;
  while (pos < len) {
#ifdef CYGWIN
    i = send(s,
	     &((char*)buf)[pos],
	     len - pos,
	     MSG_NOSIGNAL);
#elif LINUX
    i = send(s, 
	     &((char*)buf)[pos], 
	     len - pos,
	     MSG_WAITALL|MSG_NOSIGNAL); 
#else
    i = send(s,
	     &((char*)buf)[pos],
	     len - pos,
	     MSG_WAITALL);
#endif  
    if ( (i == -1) &&
	 (errno == EINTR) )
      continue; /* ingnore interrupts */
    if (i <= 0) 
      return SYSERR;    
    pos += i;
  }
  if (pos != len)
    errexit("ASSERTION failed: %u != %u\n",
	    len, pos);
  return pos;
}
	 


/* end of io.c */
